@@ -26,18 +26,19 @@ from botocore.model import ServiceModel
 from botocore.paginate import Paginator
 from botocore.utils import CachedProperty
 from botocore.utils import get_service_module_name
-from botocore.utils import switch_host_s3_accelerate
 from botocore.utils import S3RegionRedirector
-from botocore.utils import fix_s3_host
-from botocore.utils import switch_to_virtual_host_style
-from botocore.utils import S3_ACCELERATE_WHITELIST
+from botocore.utils import S3ArnParamHandler
+from botocore.utils import S3EndpointSetter
 from botocore.args import ClientArgsCreator
-from botocore.compat import urlsplit
 from botocore import UNSIGNED
 # Keep this imported.  There's pre-existing code that uses
 # "from botocore.client import Config".
 from botocore.config import Config
 from botocore.history import get_global_history_recorder
+from botocore.discovery import (
+    EndpointDiscoveryHandler, EndpointDiscoveryManager,
+    block_endpoint_discovery_required_operations
+)
 
 
 logger = logging.getLogger(__name__)
@@ -48,7 +49,8 @@ class ClientCreator(object):
     """Creates client objects for a service."""
     def __init__(self, loader, endpoint_resolver, user_agent, event_emitter,
                  retry_handler_factory, retry_config_translator,
-                 response_parser_factory=None, exceptions_factory=None):
+                 response_parser_factory=None, exceptions_factory=None,
+                 config_store=None):
         self._loader = loader
         self._endpoint_resolver = endpoint_resolver
         self._user_agent = user_agent
@@ -57,6 +59,11 @@ class ClientCreator(object):
         self._retry_config_translator = retry_config_translator
         self._response_parser_factory = response_parser_factory
         self._exceptions_factory = exceptions_factory
+        # TODO: Migrate things away from scoped_config in favor of the
+        # config_store.  The config store can pull things from both the scoped
+        # config and environment variables (and potentially more in the
+        # future).
+        self._config_store = config_store
 
     def create_client(self, service_name, region_name, is_secure=True,
                       endpoint_url=None, verify=None,
@@ -79,6 +86,9 @@ class ClientCreator(object):
         self._register_s3_events(
             service_client, endpoint_bridge, endpoint_url, client_config,
             scoped_config)
+        self._register_endpoint_discovery(
+            service_client, endpoint_url, client_config
+        )
         return service_client
 
     def create_client_class(self, service_name, api_version=None):
@@ -90,9 +100,11 @@ class ClientCreator(object):
         py_name_to_operation_name = self._create_name_mapping(service_model)
         class_attributes['_PY_TO_OP_NAME'] = py_name_to_operation_name
         bases = [BaseClient]
-        self._event_emitter.emit('creating-client-class.%s' % service_name,
-                                 class_attributes=class_attributes,
-                                 base_classes=bases)
+        service_id = service_model.service_id.hyphenize()
+        self._event_emitter.emit(
+            'creating-client-class.%s' % service_id,
+            class_attributes=class_attributes,
+            base_classes=bases)
         class_name = get_service_module_name(service_model)
         cls = type(str(class_name), tuple(bases), class_attributes)
         return cls
@@ -105,6 +117,8 @@ class ClientCreator(object):
 
     def _register_retries(self, client):
         endpoint_prefix = client.meta.service_model.endpoint_prefix
+        service_id = client.meta.service_model.service_id
+        service_event_name = service_id.hyphenize()
 
         # First, we load the entire retry config for all services,
         # then pull out just the information we need.
@@ -122,117 +136,51 @@ class ClientCreator(object):
                      client.meta.service_model.service_name)
         handler = self._retry_handler_factory.create_retry_handler(
             retry_config, endpoint_prefix)
-        unique_id = 'retry-config-%s' % endpoint_prefix
-        client.meta.events.register('needs-retry.%s' % endpoint_prefix,
-                                    handler, unique_id=unique_id)
+        unique_id = 'retry-config-%s' % service_event_name
+        client.meta.events.register(
+            'needs-retry.%s' % service_event_name, handler,
+            unique_id=unique_id
+        )
+
+    def _register_endpoint_discovery(self, client, endpoint_url, config):
+        if endpoint_url is not None:
+            # Don't register any handlers in the case of a custom endpoint url
+            return
+        # Only attach handlers if the service supports discovery
+        if client.meta.service_model.endpoint_discovery_operation is None:
+            return
+        events = client.meta.events
+        service_id = client.meta.service_model.service_id.hyphenize()
+        enabled = False
+        if config and config.endpoint_discovery_enabled is not None:
+            enabled = config.endpoint_discovery_enabled
+        elif self._config_store:
+            enabled = self._config_store.get_config_variable(
+                'endpoint_discovery_enabled')
+        if enabled:
+            manager = EndpointDiscoveryManager(client)
+            handler = EndpointDiscoveryHandler(manager)
+            handler.register(events, service_id)
+        else:
+            events.register('before-parameter-build',
+                            block_endpoint_discovery_required_operations)
 
     def _register_s3_events(self, client, endpoint_bridge, endpoint_url,
                             client_config, scoped_config):
         if client.meta.service_model.service_name != 's3':
             return
         S3RegionRedirector(endpoint_bridge, client).register()
-        self._set_s3_addressing_style(
-            endpoint_url, client.meta.config.s3, client.meta.events,
-            client.meta.partition
-        )
-        # Enable accelerate if the configuration is set to to true or the
-        # endpoint being used matches one of the accelerate endpoints.
-        if self._is_s3_accelerate(endpoint_url, client.meta.config.s3):
-            # Also make sure that the hostname gets switched to
-            # s3-accelerate.amazonaws.com
-            client.meta.events.register_first(
-                'before-sign.s3', switch_host_s3_accelerate)
-
+        S3ArnParamHandler().register(client.meta.events)
+        S3EndpointSetter(
+            endpoint_resolver=self._endpoint_resolver,
+            region=client.meta.region_name,
+            s3_config=client.meta.config.s3,
+            endpoint_url=endpoint_url,
+            partition=client.meta.partition
+        ).register(client.meta.events)
         self._set_s3_presign_signature_version(
             client.meta, client_config, scoped_config)
 
-    def _set_s3_addressing_style(self, endpoint_url, s3_config, event_emitter,
-                                 partition):
-        if s3_config is None:
-            s3_config = {}
-
-        addressing_style = self._get_s3_addressing_style(
-            endpoint_url, s3_config)
-        handler = self._get_s3_addressing_handler(
-            endpoint_url, s3_config, addressing_style, partition)
-        if handler is not None:
-            event_emitter.register('before-sign.s3', handler)
-
-    def _get_s3_addressing_style(self, endpoint_url, s3_config):
-        # Use virtual host style addressing if accelerate is enabled or if
-        # the given endpoint url is an accelerate endpoint.
-        accelerate = s3_config.get('use_accelerate_endpoint', False)
-        if accelerate or self._is_s3_accelerate(endpoint_url, s3_config):
-            return 'virtual'
-
-        # If a particular addressing style is configured, use it.
-        configured_addressing_style = s3_config.get('addressing_style')
-        if configured_addressing_style:
-            return configured_addressing_style
-
-    def _get_s3_addressing_handler(self, endpoint_url, s3_config,
-                                   addressing_style, partition):
-        # If virtual host style was configured, use it regardless of whether
-        # or not the bucket looks dns compatible.
-        if addressing_style == 'virtual':
-            logger.debug("Using S3 virtual host style addressing.")
-            return switch_to_virtual_host_style
-
-        # If path style is configured, no additional steps are needed. If
-        # endpoint_url was specified, don't default to virtual. We could
-        # potentially default provided endpoint urls to virtual hosted
-        # style, but for now it is avoided.
-        if addressing_style == 'path' or endpoint_url is not None:
-            logger.debug("Using S3 path style addressing.")
-            return None
-
-        logger.debug("Defaulting to S3 virtual host style addressing with "
-                     "path style addressing fallback.")
-
-        # For dual stack mode, we need to clear the default endpoint url in
-        # order to use the existing netloc if the bucket is dns compatible.
-        # Also, the default_endpoint_url of 's3.amazonaws.com' only works
-        # if we're in the 'aws' partition.  Anywhere else we should
-        # just use the existing netloc.
-        if s3_config.get('use_dualstack_endpoint', False) or \
-                partition != 'aws':
-            return functools.partial(
-                fix_s3_host, default_endpoint_url=None)
-
-        # By default, try to use virtual style with path fallback.
-        return fix_s3_host
-
-    def _is_s3_accelerate(self, endpoint_url, s3_config):
-        # Accelerate has been explicitly configured.
-        if s3_config is not None and s3_config.get('use_accelerate_endpoint'):
-            return True
-
-        # Accelerate mode is turned on automatically if an endpoint url is
-        # provided that matches the accelerate scheme.
-        if endpoint_url is None:
-            return False
-
-        # Accelerate is only valid for Amazon endpoints.
-        netloc = urlsplit(endpoint_url).netloc
-        if not netloc.endswith('amazonaws.com'):
-            return False
-
-        # The first part of the url should always be s3-accelerate.
-        parts = netloc.split('.')
-        if parts[0] != 's3-accelerate':
-            return False
-
-        # Url parts between 's3-accelerate' and 'amazonaws.com' which
-        # represent different url features.
-        feature_parts = parts[1:-2]
-
-        # There should be no duplicate url parts.
-        if len(feature_parts) != len(set(feature_parts)):
-            return False
-
-        # Remaining parts must all be in the whitelist.
-        return all(p in S3_ACCELERATE_WHITELIST for p in feature_parts)
-
     def _set_s3_presign_signature_version(self, client_meta,
                                           client_config, scoped_config):
         # This will return the manually configured signature version, or None
@@ -246,9 +194,13 @@ class ClientCreator(object):
         # Check to see if the region is a region that we know about. If we
         # don't know about a region, then we can safely assume it's a new
         # region that is sigv4 only, since all new S3 regions only allow sigv4.
+        # The only exception is aws-global. This is a pseudo-region for the
+        # global endpoint, we should respect the signature versions it
+        # supports, which includes v2.
         regions = self._endpoint_resolver.get_available_endpoints(
             's3', client_meta.partition)
-        if client_meta.region_name not in regions:
+        if client_meta.region_name != 'aws-global' and \
+                client_meta.region_name not in regions:
             return
 
         # If it is a region we know about, we want to default to sigv2, so here
@@ -289,7 +241,7 @@ class ClientCreator(object):
         args_creator = ClientArgsCreator(
             self._event_emitter, self._user_agent,
             self._response_parser_factory, self._loader,
-            self._exceptions_factory)
+            self._exceptions_factory, config_store=self._config_store)
         return args_creator.get_client_args(
             service_model, region_name, is_secure, endpoint_url,
             verify, credentials, scoped_config, client_config, endpoint_bridge)
@@ -351,6 +303,7 @@ class ClientEndpointBridge(object):
     utilize "us-east-1" by default if no region can be resolved."""
 
     DEFAULT_ENDPOINT = '{service}.{region}.amazonaws.com'
+    _DUALSTACK_ENABLED_SERVICES = ['s3', 's3-control']
 
     def __init__(self, endpoint_resolver, scoped_config=None,
                  client_config=None, default_endpoint=None,
@@ -382,13 +335,14 @@ class ClientEndpointBridge(object):
 
     def _create_endpoint(self, resolved, service_name, region_name,
                          endpoint_url, is_secure):
+        explicit_region = region_name is not None
         region_name, signing_region = self._pick_region_values(
             resolved, region_name, endpoint_url)
         if endpoint_url is None:
             if self._is_s3_dualstack_mode(service_name):
                 endpoint_url = self._create_dualstack_endpoint(
                     service_name, region_name,
-                    resolved['dnsSuffix'], is_secure)
+                    resolved['dnsSuffix'], is_secure, explicit_region)
             else:
                 # Use the sslCommonName over the hostname for Python 2.6 compat.
                 hostname = resolved.get('sslCommonName', resolved.get('hostname'))
@@ -404,7 +358,7 @@ class ClientEndpointBridge(object):
             signature_version=signature_version)
 
     def _is_s3_dualstack_mode(self, service_name):
-        if service_name != 's3':
+        if service_name not in self._DUALSTACK_ENABLED_SERVICES:
             return False
         # TODO: This normalization logic is duplicated from the
         # ClientArgsCreator class.  Consolidate everything to
@@ -424,7 +378,12 @@ class ClientEndpointBridge(object):
         return False
 
     def _create_dualstack_endpoint(self, service_name, region_name,
-                                   dns_suffix, is_secure):
+                                   dns_suffix, is_secure, explicit_region):
+        if not explicit_region and region_name == 'aws-global':
+            # If the region_name passed was not explicitly set, default to
+            # us-east-1 instead of the modeled default aws-global. Dualstack
+            # does not support aws-global
+            region_name = 'us-east-1'
         hostname = '{service}.dualstack.{region}.{dns_suffix}'.format(
             service=service_name, region=region_name,
             dns_suffix=dns_suffix)
@@ -518,8 +477,8 @@ class ClientEndpointBridge(object):
                 return 'v4'
             # Now just iterate over the signature versions in order until we
             # find the first one that is known to Botocore.
-            for known in AUTH_TYPE_MAPS:
-                if known in potential_versions:
+            for known in potential_versions:
+                if known in AUTH_TYPE_MAPS:
                     return known
         raise UnknownSignatureVersionError(
             signature_version=resolved.get('signatureVersions'))
@@ -553,7 +512,9 @@ class BaseClient(object):
         self._register_handlers()
 
     def __getattr__(self, item):
-        event_name = 'getattr.%s.%s' % (self._service_model.service_name, item)
+        event_name = 'getattr.%s.%s' % (
+            self._service_model.service_id.hyphenize(), item
+        )
         handler, event_response = self.meta.events.emit_until_response(
             event_name, client=self)
 
@@ -567,9 +528,11 @@ class BaseClient(object):
 
     def _register_handlers(self):
         # Register the handler required to sign requests.
-        self.meta.events.register('request-created.%s' %
-                                  self.meta.service_model.endpoint_prefix,
-                                  self._request_signer.handler)
+        service_id = self.meta.service_model.service_id.hyphenize()
+        self.meta.events.register(
+            'request-created.%s' % service_id,
+            self._request_signer.handler
+        )
 
     @property
     def _service_model(self):
@@ -595,9 +558,10 @@ class BaseClient(object):
         request_dict = self._convert_to_request_dict(
             api_params, operation_model, context=request_context)
 
+        service_id = self._service_model.service_id.hyphenize()
         handler, event_response = self.meta.events.emit_until_response(
-            'before-call.{endpoint_prefix}.{operation_name}'.format(
-                endpoint_prefix=self._service_model.endpoint_prefix,
+            'before-call.{service_id}.{operation_name}'.format(
+                service_id=service_id,
                 operation_name=operation_name),
             model=operation_model, params=request_dict,
             request_signer=self._request_signer, context=request_context)
@@ -605,12 +569,12 @@ class BaseClient(object):
         if event_response is not None:
             http, parsed_response = event_response
         else:
-            http, parsed_response = self._endpoint.make_request(
-                operation_model, request_dict)
+            http, parsed_response = self._make_request(
+                operation_model, request_dict, request_context)
 
         self.meta.events.emit(
-            'after-call.{endpoint_prefix}.{operation_name}'.format(
-                endpoint_prefix=self._service_model.endpoint_prefix,
+            'after-call.{service_id}.{operation_name}'.format(
+                service_id=service_id,
                 operation_name=operation_name),
             http_response=http, parsed=parsed_response,
             model=operation_model, context=request_context
@@ -623,12 +587,26 @@ class BaseClient(object):
         else:
             return parsed_response
 
+    def _make_request(self, operation_model, request_dict, request_context):
+        try:
+            return self._endpoint.make_request(operation_model, request_dict)
+        except Exception as e:
+            self.meta.events.emit(
+                'after-call-error.{service_id}.{operation_name}'.format(
+                    service_id=self._service_model.service_id.hyphenize(),
+                    operation_name=operation_model.name),
+                exception=e, context=request_context
+            )
+            raise
+
     def _convert_to_request_dict(self, api_params, operation_model,
                                  context=None):
         api_params = self._emit_api_params(
             api_params, operation_model, context)
         request_dict = self._serializer.serialize_to_request(
             api_params, operation_model)
+        if not self._client_config.inject_host_prefix:
+            request_dict.pop('host_prefix', None)
         prepare_request_dict(request_dict, endpoint_url=self._endpoint.host,
                              user_agent=self._client_config.user_agent,
                              context=context)
@@ -642,18 +620,19 @@ class BaseClient(object):
         # Emit an event that allows users to modify the parameters at the
         # beginning of the method. It allows handlers to modify existing
         # parameters or return a new set of parameters to use.
+        service_id = self._service_model.service_id.hyphenize()
         responses = self.meta.events.emit(
-            'provide-client-params.{endpoint_prefix}.{operation_name}'.format(
-                endpoint_prefix=self._service_model.endpoint_prefix,
+            'provide-client-params.{service_id}.{operation_name}'.format(
+                service_id=service_id,
                 operation_name=operation_name),
             params=api_params, model=operation_model, context=context)
         api_params = first_non_none_response(responses, default=api_params)
 
         event_name = (
-            'before-parameter-build.{endpoint_prefix}.{operation_name}')
+            'before-parameter-build.{service_id}.{operation_name}')
         self.meta.events.emit(
             event_name.format(
-                endpoint_prefix=self._service_model.endpoint_prefix,
+                service_id=service_id,
                 operation_name=operation_name),
             params=api_params, model=operation_model, context=context)
         return api_params
@@ -755,6 +734,15 @@ class BaseClient(object):
         return self._cache['waiter_config']
 
     def get_waiter(self, waiter_name):
+        """Returns an object that can wait for some condition.
+
+        :type waiter_name: str
+        :param waiter_name: The name of the waiter to get. See the waiters
+            section of the service docs for a list of available waiters.
+
+        :returns: The specified waiter object.
+        :rtype: botocore.waiter.Waiter
+        """
         config = self._get_waiter_config()
         if not config:
             raise ValueError("Waiter does not exist: %s" % waiter_name)
