@@ -27,13 +27,7 @@ from io import BytesIO
 from subprocess import Popen, PIPE
 
 from dateutil.tz import tzlocal
-# The unittest module got a significant overhaul
-# in 2.7, so if we're in 2.6 we can use the backported
-# version unittest2.
-if sys.version_info[:2] == (2, 6):
-    import unittest2 as unittest
-else:
-    import unittest
+import unittest
 
 from nose.tools import assert_equal
 
@@ -378,11 +372,11 @@ class RawResponse(BytesIO):
             contents = self.read()
 
 
-class ClientHTTPStubber(object):
-    def __init__(self, client, strict=True):
+class BaseHTTPStubber(object):
+    def __init__(self, obj_with_event_emitter, strict=True):
         self.reset()
         self._strict = strict
-        self._client = client
+        self._obj_with_event_emitter = obj_with_event_emitter
 
     def reset(self):
         self.requests = []
@@ -397,11 +391,15 @@ class ClientHTTPStubber(object):
         response = AWSResponse(url, status, headers, raw)
         self.responses.append(response)
 
+    @property
+    def _events(self):
+        raise NotImplementedError('_events')
+
     def start(self):
-        self._client.meta.events.register('before-send', self)
+        self._events.register('before-send', self)
 
     def stop(self):
-        self._client.meta.events.unregister('before-send', self)
+        self._events.unregister('before-send', self)
 
     def __enter__(self):
         self.start()
@@ -424,6 +422,18 @@ class ClientHTTPStubber(object):
             return None
 
 
+class ClientHTTPStubber(BaseHTTPStubber):
+    @property
+    def _events(self):
+        return self._obj_with_event_emitter.meta.events
+
+
+class SessionHTTPStubber(BaseHTTPStubber):
+    @property
+    def _events(self):
+        return self._obj_with_event_emitter.get_component('event_emitter')
+
+
 class ConsistencyWaiterException(Exception):
     pass
 
