@@ -25,18 +25,18 @@ from dateutil.tz import tzlocal
 from botocore.exceptions import CredentialRetrievalError
 
 from tests import unittest, IntegerRefresher, BaseEnvVar, random_chars
-from tests import ClientHTTPStubber
-from tests import temporary_file
+from tests import temporary_file, StubbedSession, SessionHTTPStubber
 from botocore.credentials import EnvProvider, ContainerProvider
 from botocore.credentials import InstanceMetadataProvider
 from botocore.credentials import Credentials, ReadOnlyCredentials
-from botocore.credentials import AssumeRoleProvider
+from botocore.credentials import AssumeRoleProvider, ProfileProviderBuilder
 from botocore.credentials import CanonicalNameCredentialSourcer
 from botocore.credentials import DeferredRefreshableCredentials
 from botocore.credentials import create_credential_resolver
 from botocore.session import Session
 from botocore.exceptions import InvalidConfigError, InfiniteLoopConfigError
 from botocore.stub import Stubber
+from botocore.awsrequest import AWSResponse
 
 
 class TestCredentialRefreshRaces(unittest.TestCase):
@@ -142,76 +142,21 @@ class TestCredentialRefreshRaces(unittest.TestCase):
         self.assert_non_none_retrieved_credentials(_run_in_thread)
 
 
-class TestAssumeRole(BaseEnvVar):
+class BaseAssumeRoleTest(BaseEnvVar):
     def setUp(self):
-        super(TestAssumeRole, self).setUp()
+        super(BaseAssumeRoleTest, self).setUp()
         self.tempdir = tempfile.mkdtemp()
         self.config_file = os.path.join(self.tempdir, 'config')
         self.environ['AWS_CONFIG_FILE'] = self.config_file
         self.environ['AWS_SHARED_CREDENTIALS_FILE'] = str(uuid.uuid4())
-        self.environ['AWS_ACCESS_KEY_ID'] = 'access_key'
-        self.environ['AWS_SECRET_ACCESS_KEY'] = 'secret_key'
-
-        self.metadata_provider = self.mock_provider(InstanceMetadataProvider)
-        self.env_provider = self.mock_provider(EnvProvider)
-        self.container_provider = self.mock_provider(ContainerProvider)
-        self.actual_client_region = None
-
-    def mock_provider(self, provider_cls):
-        mock_instance = mock.Mock(spec=provider_cls)
-        mock_instance.load.return_value = None
-        mock_instance.METHOD = provider_cls.METHOD
-        mock_instance.CANONICAL_NAME = provider_cls.CANONICAL_NAME
-        return mock_instance
 
     def tearDown(self):
         shutil.rmtree(self.tempdir)
+        super(BaseAssumeRoleTest, self).tearDown()
 
-    def create_session(self, profile=None):
-        session = Session(profile=profile)
-
-        # We have to set bogus credentials here or otherwise we'll trigger
-        # an early credential chain resolution.
-        sts = session.create_client(
-            'sts',
-            aws_access_key_id='spam',
-            aws_secret_access_key='eggs',
-        )
-        stubber = Stubber(sts)
-        stubber.activate()
-        assume_role_provider = AssumeRoleProvider(
-            load_config=lambda: session.full_config,
-            client_creator=lambda *args, **kwargs: sts,
-            cache={},
-            profile_name=profile,
-            credential_sourcer=CanonicalNameCredentialSourcer([
-                self.env_provider, self.container_provider,
-                self.metadata_provider
-            ])
-        )
-
-        component_name = 'credential_provider'
-        resolver = session.get_component(component_name)
-        available_methods = [p.METHOD for p in resolver.providers]
-        replacements = {
-            'env': self.env_provider,
-            'iam-role': self.metadata_provider,
-            'container-role': self.container_provider,
-            'assume-role': assume_role_provider
-        }
-        for name, provider in replacements.items():
-            try:
-                index = available_methods.index(name)
-            except ValueError:
-                # The provider isn't in the session
-                continue
-
-            resolver.providers[index] = provider
-
-        session.register_component(
-            'credential_provider', resolver
-        )
-        return session, stubber
+    def some_future_time(self):
+        timeobj = datetime.now(tzlocal())
+        return timeobj + timedelta(hours=24)
 
     def create_assume_role_response(self, credentials, expiration=None):
         if expiration is None:
@@ -239,14 +184,6 @@ class TestAssumeRole(BaseEnvVar):
             'fake-%s' % random_chars(45)
         )
 
-    def some_future_time(self):
-        timeobj = datetime.now(tzlocal())
-        return timeobj + timedelta(hours=24)
-
-    def write_config(self, config):
-        with open(self.config_file, 'w') as f:
-            f.write(config)
-
     def assert_creds_equal(self, c1, c2):
         c1_frozen = c1
         if not isinstance(c1_frozen, ReadOnlyCredentials):
@@ -256,6 +193,86 @@ class TestAssumeRole(BaseEnvVar):
             c2_frozen = c2.get_frozen_credentials()
         self.assertEqual(c1_frozen, c2_frozen)
 
+    def write_config(self, config):
+        with open(self.config_file, 'w') as f:
+            f.write(config)
+
+
+class TestAssumeRole(BaseAssumeRoleTest):
+    def setUp(self):
+        super(TestAssumeRole, self).setUp()
+        self.environ['AWS_ACCESS_KEY_ID'] = 'access_key'
+        self.environ['AWS_SECRET_ACCESS_KEY'] = 'secret_key'
+
+        self.metadata_provider = self.mock_provider(InstanceMetadataProvider)
+        self.env_provider = self.mock_provider(EnvProvider)
+        self.container_provider = self.mock_provider(ContainerProvider)
+        self.mock_client_creator = mock.Mock(spec=Session.create_client)
+        self.actual_client_region = None
+
+        current_dir = os.path.dirname(os.path.abspath(__file__))
+        credential_process = os.path.join(
+            current_dir, 'utils', 'credentialprocess.py'
+        )
+        self.credential_process = '%s %s' % (
+            sys.executable, credential_process
+        )
+
+    def mock_provider(self, provider_cls):
+        mock_instance = mock.Mock(spec=provider_cls)
+        mock_instance.load.return_value = None
+        mock_instance.METHOD = provider_cls.METHOD
+        mock_instance.CANONICAL_NAME = provider_cls.CANONICAL_NAME
+        return mock_instance
+
+    def create_session(self, profile=None):
+        session = StubbedSession(profile=profile)
+
+        # We have to set bogus credentials here or otherwise we'll trigger
+        # an early credential chain resolution.
+        sts = session.create_client(
+            'sts',
+            aws_access_key_id='spam',
+            aws_secret_access_key='eggs',
+        )
+        self.mock_client_creator.return_value = sts
+        assume_role_provider = AssumeRoleProvider(
+            load_config=lambda: session.full_config,
+            client_creator=self.mock_client_creator,
+            cache={},
+            profile_name=profile,
+            credential_sourcer=CanonicalNameCredentialSourcer([
+                self.env_provider, self.container_provider,
+                self.metadata_provider
+            ]),
+            profile_provider_builder=ProfileProviderBuilder(session),
+        )
+        stubber = session.stub('sts')
+        stubber.activate()
+
+        component_name = 'credential_provider'
+        resolver = session.get_component(component_name)
+        available_methods = [p.METHOD for p in resolver.providers]
+        replacements = {
+            'env': self.env_provider,
+            'iam-role': self.metadata_provider,
+            'container-role': self.container_provider,
+            'assume-role': assume_role_provider
+        }
+        for name, provider in replacements.items():
+            try:
+                index = available_methods.index(name)
+            except ValueError:
+                # The provider isn't in the session
+                continue
+
+            resolver.providers[index] = provider
+
+        session.register_component(
+            'credential_provider', resolver
+        )
+        return session, stubber
+
     def test_assume_role(self):
         config = (
             '[profile A]\n'
@@ -360,13 +377,13 @@ class TestAssumeRole(BaseEnvVar):
             'role_arn = arn:aws:iam::123456789:role/RoleA\n'
             'source_profile = B\n'
             '[profile B]\n'
-            'credential_process = command\n'
+            'region = us-west-2\n'
         )
         self.write_config(config)
 
         with self.assertRaises(InvalidConfigError):
             session, _ = self.create_session(profile='A')
-            session.get_credentials()
+            session.get_credentials().get_frozen_credentials()
 
     def test_recursive_assume_role(self):
         config = (
@@ -432,6 +449,118 @@ class TestAssumeRole(BaseEnvVar):
             session, _ = self.create_session(profile='A')
             session.get_credentials()
 
+    def test_process_source_profile(self):
+        config = (
+            '[profile A]\n'
+            'role_arn = arn:aws:iam::123456789:role/RoleA\n'
+            'source_profile = B\n'
+            '[profile B]\n'
+            'credential_process = %s\n' % self.credential_process
+        )
+        self.write_config(config)
+
+        expected_creds = self.create_random_credentials()
+        response = self.create_assume_role_response(expected_creds)
+        session, stubber = self.create_session(profile='A')
+        stubber.add_response('assume_role', response)
+
+        actual_creds = session.get_credentials()
+        self.assert_creds_equal(actual_creds, expected_creds)
+        stubber.assert_no_pending_responses()
+        # Assert that the client was created with the credentials from the
+        # credential process.
+        self.assertEqual(self.mock_client_creator.call_count, 1)
+        _, kwargs = self.mock_client_creator.call_args_list[0]
+        expected_kwargs = {
+            'aws_access_key_id': 'spam',
+            'aws_secret_access_key': 'eggs',
+            'aws_session_token': None,
+        }
+        self.assertEqual(kwargs, expected_kwargs)
+
+    def test_web_identity_source_profile(self):
+        token_path = os.path.join(self.tempdir, 'token')
+        with open(token_path, 'w') as token_file:
+            token_file.write('a.token')
+        config = (
+            '[profile A]\n'
+            'role_arn = arn:aws:iam::123456789:role/RoleA\n'
+            'source_profile = B\n'
+            '[profile B]\n'
+            'role_arn = arn:aws:iam::123456789:role/RoleB\n'
+            'web_identity_token_file = %s\n' % token_path
+        )
+        self.write_config(config)
+
+        session, stubber = self.create_session(profile='A')
+
+        identity_creds = self.create_random_credentials()
+        identity_response = self.create_assume_role_response(identity_creds)
+        stubber.add_response(
+            'assume_role_with_web_identity',
+            identity_response,
+        )
+
+        expected_creds = self.create_random_credentials()
+        assume_role_response = self.create_assume_role_response(expected_creds)
+        stubber.add_response('assume_role', assume_role_response)
+
+        actual_creds = session.get_credentials()
+        self.assert_creds_equal(actual_creds, expected_creds)
+        stubber.assert_no_pending_responses()
+        # Assert that the client was created with the credentials from the
+        # assume role with web identity call.
+        self.assertEqual(self.mock_client_creator.call_count, 1)
+        _, kwargs = self.mock_client_creator.call_args_list[0]
+        expected_kwargs = {
+            'aws_access_key_id': identity_creds.access_key,
+            'aws_secret_access_key': identity_creds.secret_key,
+            'aws_session_token': identity_creds.token,
+        }
+        self.assertEqual(kwargs, expected_kwargs)
+
+    def test_web_identity_source_profile_ignores_env_vars(self):
+        token_path = os.path.join(self.tempdir, 'token')
+        with open(token_path, 'w') as token_file:
+            token_file.write('a.token')
+        self.environ['AWS_ROLE_ARN'] = 'arn:aws:iam::123456789:role/RoleB'
+        config = (
+            '[profile A]\n'
+            'role_arn = arn:aws:iam::123456789:role/RoleA\n'
+            'source_profile = B\n'
+            '[profile B]\n'
+            'web_identity_token_file = %s\n' % token_path
+        )
+        self.write_config(config)
+
+        session, _ = self.create_session(profile='A')
+        # The config is split between the profile and the env, we
+        # should only be looking at the profile so this should raise
+        # a configuration error.
+        with self.assertRaises(InvalidConfigError):
+            session.get_credentials()
+
+    def test_web_identity_credential_source_ignores_env_vars(self):
+        token_path = os.path.join(self.tempdir, 'token')
+        with open(token_path, 'w') as token_file:
+            token_file.write('a.token')
+        self.environ['AWS_ROLE_ARN'] = 'arn:aws:iam::123456789:role/RoleB'
+        self.environ['AWS_WEB_IDENTITY_TOKEN_FILE'] = token_path
+        config = (
+            '[profile A]\n'
+            'role_arn = arn:aws:iam::123456789:role/RoleA\n'
+            'credential_source = Environment\n'
+        )
+        self.write_config(config)
+
+        session, _ = self.create_session(profile='A')
+        # We should not get credentials from web-identity configured in the
+        # environment when the Environment credential_source is set.
+        # There are no Environment credentials, so this should raise a
+        # retrieval error.
+        with self.assertRaises(CredentialRetrievalError):
+            session.get_credentials()
+
     def test_self_referential_profile(self):
         config = (
             '[profile A]\n'
@@ -451,7 +580,6 @@ class TestAssumeRole(BaseEnvVar):
         self.assert_creds_equal(actual_creds, expected_creds)
         stubber.assert_no_pending_responses()
 
-
     def create_stubbed_sts_client(self, session):
         expected_creds = self.create_random_credentials()
         _original_create_client = session.create_client
@@ -492,6 +620,88 @@ class TestAssumeRole(BaseEnvVar):
         self.assertEqual(self.actual_client_region, 'cn-north-1')
 
 
+class TestAssumeRoleWithWebIdentity(BaseAssumeRoleTest):
+    def setUp(self):
+        super(TestAssumeRoleWithWebIdentity, self).setUp()
+        self.token_file = os.path.join(self.tempdir, 'token.jwt')
+        self.write_token('totally.a.token')
+
+    def write_token(self, token, path=None):
+        if path is None:
+            path = self.token_file
+        with open(path, 'w') as f:
+            f.write(token)
+
+    def assert_session_credentials(self, expected_params, **kwargs):
+        expected_creds = self.create_random_credentials()
+        response = self.create_assume_role_response(expected_creds)
+        session = StubbedSession(**kwargs)
+        stubber = session.stub('sts')
+        stubber.add_response(
+            'assume_role_with_web_identity',
+            response,
+            expected_params
+        )
+        stubber.activate()
+        actual_creds = session.get_credentials()
+        self.assert_creds_equal(actual_creds, expected_creds)
+        stubber.assert_no_pending_responses()
+
+    def test_assume_role(self):
+        config = (
+            '[profile A]\n'
+            'role_arn = arn:aws:iam::123456789:role/RoleA\n'
+            'role_session_name = sname\n'
+            'web_identity_token_file = %s\n'
+        ) % self.token_file
+        self.write_config(config)
+        expected_params = {
+            'RoleArn': 'arn:aws:iam::123456789:role/RoleA',
+            'RoleSessionName': 'sname',
+            'WebIdentityToken': 'totally.a.token',
+        }
+        self.assert_session_credentials(expected_params, profile='A')
+
+    def test_assume_role_env_vars(self):
+        config = (
+            '[profile B]\n'
+            'region = us-west-2\n'
+        )
+        self.write_config(config)
+        self.environ['AWS_ROLE_ARN'] = 'arn:aws:iam::123456789:role/RoleB'
+        self.environ['AWS_WEB_IDENTITY_TOKEN_FILE'] = self.token_file
+        self.environ['AWS_ROLE_SESSION_NAME'] = 'bname'
+
+        expected_params = {
+            'RoleArn': 'arn:aws:iam::123456789:role/RoleB',
+            'RoleSessionName': 'bname',
+            'WebIdentityToken': 'totally.a.token',
+        }
+        self.assert_session_credentials(expected_params)
+
+    def test_assume_role_env_vars_do_not_take_precedence(self):
+        config = (
+            '[profile A]\n'
+            'role_arn = arn:aws:iam::123456789:role/RoleA\n'
+            'role_session_name = aname\n'
+            'web_identity_token_file = %s\n'
+        ) % self.token_file
+        self.write_config(config)
+
+        different_token = os.path.join(self.tempdir, str(uuid.uuid4()))
+        self.write_token('totally.different.token', path=different_token)
+        self.environ['AWS_ROLE_ARN'] = 'arn:aws:iam::123456789:role/RoleC'
+        self.environ['AWS_WEB_IDENTITY_TOKEN_FILE'] = different_token
+        self.environ['AWS_ROLE_SESSION_NAME'] = 'cname'
+
+        expected_params = {
+            'RoleArn': 'arn:aws:iam::123456789:role/RoleA',
+            'RoleSessionName': 'aname',
+            'WebIdentityToken': 'totally.a.token',
+        }
+        self.assert_session_credentials(expected_params, profile='A')
+
+
 class TestProcessProvider(unittest.TestCase):
     def setUp(self):
         current_dir = os.path.dirname(os.path.abspath(__file__))
@@ -552,3 +762,94 @@ class TestProcessProvider(unittest.TestCase):
             reg = r"(?s)^((?!b').)*$"
             with self.assertRaisesRegexp(CredentialRetrievalError, reg):
                 session.get_credentials()
+
+
+class TestSTSRegional(BaseAssumeRoleTest):
+    def add_assume_role_http_response(self, stubber):
+        stubber.add_response(
+            body=self._get_assume_role_body('AssumeRole'))
+
+    def add_assume_role_with_web_identity_http_response(self, stubber):
+        stubber.add_response(
+            body=self._get_assume_role_body('AssumeRoleWithWebIdentity'))
+
+    def _get_assume_role_body(self, method_name):
+        expiration = self.some_future_time()
+        body = (
+            '<{method_name}Response>'
+            '  <{method_name}Result>'
+            '    <AssumedRoleUser>'
+            '      <Arn>arn:aws:sts::0123456:user</Arn>'
+            '      <AssumedRoleId>AKID:mysession-1567020004</AssumedRoleId>'
+            '    </AssumedRoleUser>'
+            '    <Credentials>'
+            '      <AccessKeyId>AccessKey</AccessKeyId>'
+            '      <SecretAccessKey>SecretKey</SecretAccessKey>'
+            '      <SessionToken>SessionToken</SessionToken>'
+            '      <Expiration>{expiration}</Expiration>'
+            '    </Credentials>'
+            '  </{method_name}Result>'
+            '</{method_name}Response>'
+        ).format(method_name=method_name, expiration=expiration)
+        return body.encode('utf-8')
+
+    def make_stubbed_client_call_to_region(self, session, stubber, region):
+        ec2 = session.create_client('ec2', region_name=region)
+        stubber.add_response(body=b'<DescribeRegionsResponse/>')
+        ec2.describe_regions()
+
+    def test_assume_role_uses_same_region_as_client(self):
+        config = (
+            '[profile A]\n'
+            'sts_regional_endpoints = regional\n'
+            'role_arn = arn:aws:iam::123456789:role/RoleA\n'
+            'source_profile = B\n\n'
+            '[profile B]\n'
+            'aws_access_key_id = abc123\n'
+            'aws_secret_access_key = def456\n'
+        )
+        self.write_config(config)
+
+        session = Session(profile='A')
+        with SessionHTTPStubber(session) as stubber:
+            self.add_assume_role_http_response(stubber)
+            # Make an arbitrary client and API call as we are really only
+            # looking to make sure the STS assume role call uses the correct
+            # endpoint.
+            self.make_stubbed_client_call_to_region(
+                session, stubber, 'us-west-2')
+            self.assertEqual(
+                stubber.requests[0].url,
+                'https://sts.us-west-2.amazonaws.com/'
+            )
+
+    def test_assume_role_web_identity_uses_same_region_as_client(self):
+        token_file = os.path.join(self.tempdir, 'token.jwt')
+        with open(token_file, 'w') as f:
+            f.write('some-token')
+        config = (
+            '[profile A]\n'
+            'sts_regional_endpoints = regional\n'
+            'role_arn = arn:aws:iam::123456789:role/RoleA\n'
+            'web_identity_token_file = %s\n'
+            'source_profile = B\n\n'
+            '[profile B]\n'
+            'aws_access_key_id = abc123\n'
+            'aws_secret_access_key = def456\n' % token_file
+        )
+        self.write_config(config)
+        # Make an arbitrary client and API call as we are really only
+        # looking to make sure the STS assume role call uses the correct
+        # endpoint.
+        session = Session(profile='A')
+        with SessionHTTPStubber(session) as stubber:
+            self.add_assume_role_with_web_identity_http_response(stubber)
+            # Make an arbitrary client and API call as we are really only
+            # looking to make sure the STS assume role call uses the correct
+            # endpoint.
+            self.make_stubbed_client_call_to_region(
+                session, stubber, 'us-west-2')
+            self.assertEqual(
+                stubber.requests[0].url,
+                'https://sts.us-west-2.amazonaws.com/'
+            )
