@@ -360,20 +360,6 @@ class TestS3SigV4Auth(BaseTestWithFixedDate):
     def test_blacklist_headers(self):
         self._test_blacklist_header('user-agent', 'botocore/1.4.11')
 
-    def test_context_sets_signing_region(self):
-        original_signing_region = 'eu-central-1'
-        new_signing_region = 'us-west-2'
-        self.auth.add_auth(self.request)
-        auth = self.request.headers['Authorization']
-        self.assertIn(original_signing_region, auth)
-        self.assertNotIn(new_signing_region, auth)
-
-        self.request.context = {'signing': {'region': new_signing_region}}
-        self.auth.add_auth(self.request)
-        auth = self.request.headers['Authorization']
-        self.assertIn(new_signing_region, auth)
-        self.assertNotIn(original_signing_region, auth)
-
     def test_uses_sha256_if_config_value_is_true(self):
         self.client_config.s3['payload_signing_enabled'] = True
         self.auth.add_auth(self.request)
