@@ -102,7 +102,7 @@ class TestURLLib3Session(unittest.TestCase):
             url=url,
             body=body,
             headers=headers,
-            retries=False,
+            retries=ANY,
             assert_same_host=False,
             preload_content=False,
             decode_content=False,
